from ast import fix_missing_locations
from shaderdef.ast_util import (make_assign,
                                make_self_attr_load,
                                make_self_attr_store,
                                rename_function)
from shaderdef.attr_rename import rename_attributes
from shaderdef.find_deps import find_deps
from shaderdef.find_method import find_method_ast
from shaderdef.unselfify import unselfify
from shaderdef.py_to_glsl import py_to_glsl


def make_prefix(name):
    parts = name.split('_')
    return ''.join(part[0] for part in parts) + '_'


class Stage(object):
    def __init__(self, obj, func_name):
        self.name = func_name
        self.ast_root = find_method_ast(obj, func_name)
        self.input_prefix = ''
        self.output_prefix = make_prefix(self.name)

    def find_deps(self):
        return find_deps(self.ast_root)

    def provide_deps(self, next_stage):
        for dep in next_stage.find_deps().inputs:
            if dep not in self.find_deps().outputs:
                dst = make_self_attr_store(dep)
                src = make_self_attr_load(dep)
                assign = make_assign(dst, src)
                self.ast_root.body.append(assign)

        fix_missing_locations(self.ast_root)

    def required_uniforms(self, all_uniforms):
        for link in self.find_deps().inputs:
            unif = all_uniforms.get(link)
            if unif is not None:
                yield link, unif

    def load_names(self, external_links):
        names = {}
        for link in self.find_deps().inputs:
            if link not in external_links.uniforms:
                names[link] = self.input_prefix + link
        return names

    # TODO(nicholasbishop): de-dup
    def store_names(self, external_links):
        names = {}
        for link in self.find_deps().outputs:
            # Don't prefix uniforms, external outputs, or builtins
            if link not in external_links.uniforms and \
               link not in external_links.frag_outputs and \
               not link.startswith('gl_'):
                names[link] = self.output_prefix + link
        return names

    def to_glsl(self, external_links, library):
        lines = []
        for link, unif in self.required_uniforms(external_links.uniforms):
            lines.append(unif.glsl_decl(link))

        # TODO
        if self.name == 'vert_shader':
            for link, attr in external_links.attributes.items():
                # TODO: layout=
                lines.append(attr.glsl_decl(link))

        # TODO
        if self.name == 'frag_shader':
            for link, fout in external_links.frag_outputs.items():
                lines.append(fout.glsl_decl(link))

        # TODO
        for func_name in self.find_deps().calls:
            auxfunc = find_method_ast(library, func_name)
            lines += py_to_glsl(auxfunc)

        rename_function(self.ast_root, 'main')
        ast_root = rename_attributes(
            self.ast_root,
            load_names=self.load_names(external_links),
            store_names=self.store_names(external_links))
        ast_root = unselfify(ast_root)
        lines += py_to_glsl(ast_root)
        return '\n'.join(lines)



# pylint: disable=missing-docstring

import ast
from unittest import TestCase

from shaderdef.attr_rename import rename_attributes
from shaderdef.find_deps import find_deps
from shaderdef.find_method import find_method_ast
from shaderdef.py_to_glsl import py_to_glsl
from shaderdef.unselfify import unselfify


class AstTestCase(TestCase):
    def assertEqual(self, left, right, msg=None):
        if isinstance(left, ast.AST) and isinstance(right, ast.AST):
            left = ast.dump(left)
            right = ast.dump(right)
        return super(AstTestCase, self).assertEqual(left, right, msg)


class TestUnselfify(AstTestCase):
    def test_load(self):
        root = unselfify(ast.parse('var = self.value'))
        expected = ast.parse('var = value')
        self.assertEqual(root, expected)

    def test_store(self):
        root = unselfify(ast.parse('self.var = value'))
        expected = ast.parse('var = value')
        self.assertEqual(root, expected)

    def test_nonself(self):
        root = unselfify(ast.parse('foo.var = value'))
        expected = ast.parse('foo.var = value')
        self.assertEqual(root, expected)


class SimpleClass(object):
    def my_method(self):
        pass


class TestFindMethod(TestCase):
    def test_find_method(self):
        self.assertIsNot(find_method_ast(SimpleClass, 'my_method'), None)

    def test_method_not_found(self):
        with self.assertRaises(KeyError):
            find_method_ast(SimpleClass, 'bad_name')


class TestAttrRename(AstTestCase):
    def test_load(self):
        root = rename_attributes(ast.parse('var = self.value1'),
                                 load_names={'value1': 'value2'})
        expected = ast.parse('var = self.value2')
        self.assertEqual(root, expected)

    def test_store(self):
        root = rename_attributes(ast.parse('self.var1 = value'),
                                 store_names={'var1': 'var2'})
        expected = ast.parse('self.var2 = value')
        self.assertEqual(root, expected)

    def test_load_and_store(self):
        root = rename_attributes(ast.parse('self.var1 = self.value1'),
                                 load_names={'value1': 'value2'},
                                 store_names={'var1': 'var2'})
        expected = ast.parse('self.var2 = self.value2')
        self.assertEqual(root, expected)

    def test_load_and_store_same_key(self):
        root = rename_attributes(ast.parse('self.foo1 = self.foo1'),
                                 load_names={'foo1': 'load'},
                                 store_names={'foo1': 'store'})
        expected = ast.parse('self.store = self.load')
        self.assertEqual(root, expected)

    def test_ignore_function_call(self):
        root = rename_attributes(ast.parse('self.myfunc(1)'),
                                 load_names={'myfunc': 'load'},
                                 store_names={'myfunc': 'store'})
        expected = ast.parse('self.myfunc(1)')
        self.assertEqual(root, expected)


class TestFindDeps(TestCase):
    def test_find_input(self):
        deps = find_deps(ast.parse('var = self.value'))
        self.assertEqual(deps.inputs, set(['value']))
        self.assertEqual(deps.outputs, set())

    def test_find_output(self):
        deps = find_deps(ast.parse('self.var = value'))
        self.assertEqual(deps.inputs, set())
        self.assertEqual(deps.outputs, set(['var']))

    def test_find_both(self):
        deps = find_deps(ast.parse('self.var = self.value'))
        self.assertEqual(deps.inputs, set(['value']))
        self.assertEqual(deps.outputs, set(['var']))


class TestPyToGlsl(TestCase):
    def test_empty_function(self):
        root = ast.parse('def myFunc(): pass')
        code = ''.join(py_to_glsl(root))
        self.assertEqual(code, 'void myFunc() {}')

    def test_assign(self):
        root = ast.parse('a = b')
        code = ''.join(py_to_glsl(root))
        self.assertEqual(code, 'a = b;')

    def test_binop(self):
        root = ast.parse('a - b * c / d')
        code = ''.join(py_to_glsl(root))
        self.assertEqual(code, '(a - ((b * c) / d))')



from unittest import TestCase

from shaderdef.glsl_types import FragOutput, vec4
from shaderdef.shader import ShaderDef


class MyMaterial(object):
    def __init__(self):
        self.color = FragOutput(vec4)

    def vert_shader(self):
        pass

    def frag_shader(self):
        self.color = vec4(1.0, 0.0, 0.0, 1.0)


def deindent(text):
    lines = text.splitlines()
    out_lines = []
    for line in lines:
        out_lines.append(line.lstrip())
    return ''.join(out_lines)


class TestShader(TestCase):
    def test_simple(self):
        shader = ShaderDef(MyMaterial())
        shader.translate()
        self.assertEqual(deindent(shader.vert_shader),
                         'void main() {}')
        self.assertEqual(deindent(shader.frag_shader),
                         'out vec4 color;'
                         'void main() {color = vec4(1.0, 0.0, 0.0, 1.0);}')



from unittest import TestCase

from shaderdef.glsl_types import Attribute, FragOutput, Uniform, vec4
from shaderdef.material import find_external_links


class TestFindExternalLinks(TestCase):
    def setUp(self):
        class MyMaterial(object):
            def __init__(self):
                self.myattr = Attribute(vec4)
                self.myout = FragOutput(vec4)
                self.myunif = Uniform(vec4)

        self.material = MyMaterial()

    def test_links(self):
        links = find_external_links(self.material)
        self.assertEqual(links.attributes, {'myattr': Attribute(vec4)})
        self.assertEqual(links.frag_outputs, {'myout': FragOutput(vec4)})
        self.assertEqual(links.uniforms, {'myunif': Uniform(vec4)})



# pylint: disable=missing-docstring

import ast

class _UnselfifyTransformer(ast.NodeTransformer):
    # pylint: disable=invalid-name,no-self-use
    def visit_Attribute(self, node):
        if node.value.id == 'self':
            return ast.Name(id=node.attr, ctx=node.ctx)
        else:
            return node


def unselfify(root):
    """Modify an AST so that "self.some_attr" becomes "some_attr"."""
    return _UnselfifyTransformer().visit(root)



from collections import OrderedDict

from shaderdef.glsl_types import Attribute, FragOutput, Uniform
from shaderdef.stage import Stage


class Links(object):
    def __init__(self):
        self.attributes = OrderedDict()
        self.frag_outputs = OrderedDict()
        self.uniforms = OrderedDict()


def find_external_links(material):
    links = Links()
    for key in dir(material):
        val = getattr(material, key)
        if isinstance(val, Attribute):
            links.attributes[key] = val
        elif isinstance(val, FragOutput):
            links.frag_outputs[key] = val
        elif isinstance(val, Uniform):
            links.uniforms[key] = val
    return links


def create_stages(material):
    # TODO
    stage_names = ('vert_shader', 'frag_shader')
    for name in stage_names:
        yield Stage(material.__class__, name)



import ast


def make_assign(dst, src):
    return ast.Assign(targets=[dst], value=src)


def make_self_load():
    return ast.Name(ctx=ast.Load(), id='self')


def make_self_attr_load(attr):
    obj = make_self_load()
    return ast.Attribute(attr=attr, ctx=ast.Load(), value=obj)


def make_self_attr_store(attr):
    obj = make_self_load()
    return ast.Attribute(attr=attr, ctx=ast.Store(), value=obj)


def append_to_function_body(func_node, new_node):
    ast.fix_missing_locations(new_node)
    func_node.body.append(new_node)


def rename_function(func_node, new_name):
    if not isinstance(func_node, ast.FunctionDef):
        raise TypeError('input must be an ast.FunctionDef', func_node)
    func_node.name = new_name



import ast


class Code(object):
    def __init__(self, initial_line=None):
        self.lines = []
        if initial_line is not None:
            self.lines.append(initial_line)
        self._indent_string = '    '

    def __call__(self, line):
        self.lines.append(line)

    def indent(self, code):
        for line in code.lines:
            self.lines.append(self._indent_string + line)

    def one(self):
        if len(self.lines) != 1:
            raise ValueError('expected exactly one line', self)
        return self.lines[0]


class AstToGlsl(ast.NodeVisitor):
    # pylint: disable=invalid-name
    def visit_Module(self, node):
        if len(node.body) != 1:
            raise NotImplementedError()
        child = node.body[0]
        return self.visit(child)

    def visit_FunctionDef(self, node):
        if getattr(node, 'returns', None) is None:
            return_type = 'void'
        else:
            return_type = self.visit(node.returns).one()
        params = node.args.args[1:]  # skip self
        params = (self.visit(param).one() for param in params)
        code = Code()
        code('{} {}({}) {{'.format(return_type, node.name,
                                   ', '.join(params)))
        for child in node.body:
            code.indent(self.visit(child))
        code('}')
        return code

    def visit_Pass(self, _):
        # pylint: disable=no-self-use
        return Code()

    def visit_arg(self, node):
        gtype = self.visit(node.annotation).one()
        return Code('{} {}'.format(gtype, node.arg))

    def visit_Name(self, node):
        # pylint: disable=no-self-use
        return Code(node.id)

    def visit_Attribute(self, node):
        return Code('{}.{}'.format(self.visit(node.value).one(), node.attr))

    def visit_Assign(self, node):
        if len(node.targets) != 1:
            raise ValueError('multiple assignment targets not allowed', node)
        target = node.targets[0]
        return Code('{} = {};'.format(self.visit(target).one(),
                                      self.visit(node.value).one()))

    def visit_Num(self, node):
        # pylint: disable=no-self-use
        return Code(str(node.n))

    def visit_Call(self, node):
        args = (self.visit(arg).one() for arg in node.args)
        name = self.visit(node.func).one()
        return Code('{}({})'.format(name, ', '.join(args)))

    def visit_Return(self, node):
        return Code('return {};'.format(self.visit(node.value).one()))

    def visit_Expr(self, node):
        return self.visit(node.value)

    def visit_BinOp(self, node):
        # TODO(nicholasbishop): FloorDiv, Pow, LShift, RShift,
        # BitOr, BitXor, BitAnd
        ops = {
            ast.Add: '+',
            ast.Sub: '-',
            ast.Mult: '*',
            ast.Div: '/',
            ast.Mod: '%',
        }
        if hasattr(ast, 'MatMult'):
            ops[ast.MatMult] = '*'
        return Code('({} {} {})'.format(
            self.visit(node.left).one(),
            ops[node.op.__class__],
            self.visit(node.right).one(),
        ))


def py_to_glsl(root):
    """Translate Python AST into GLSL code.

    root: an ast.FunctionDef object

    Return a list of strings, where each string is a line of GLSL
    code.
    """
    atg = AstToGlsl()
    code = atg.visit(root)
    return code.lines



# pylint: disable=missing-docstring

import ast
from inspect import getsource

class _FindMethodVisitor(ast.NodeVisitor):
    # pylint: disable=invalid-name
    def __init__(self, method_name):
        super(_FindMethodVisitor, self).__init__()
        self._method_name = method_name
        self._method = None

    @property
    def method(self):
        return self._method

    def visit_FunctionDef(self, node):
        if node.name == self._method_name:
            self._method = node


def find_method_ast(cls, method_name):
    """Get the AST for a method in a class.

    A KeyError is raised if the method is not found.

    cls: class type (not a string)

    method_name: method to search for (string)
    """
    src = getsource(cls)
    node = ast.parse(src)

    finder = _FindMethodVisitor(method_name)
    finder.visit(node)

    if finder.method is None:
        raise KeyError('method not found', method_name)
    return finder.method



import ast

class _FindDepsVisitor(ast.NodeVisitor):
    # pylint: disable=invalid-name
    def __init__(self):
        super(_FindDepsVisitor, self).__init__()
        self.inputs = set()
        self.outputs = set()
        self.calls = set()

    def visit_Call(self, node):
        if isinstance(node.func, ast.Attribute):
            self.calls.add(node.func.attr)
        else:
            self.generic_visit(node)

    def visit_Attribute(self, node):
        if node.value.id == 'self':
            if isinstance(node.ctx, ast.Load):
                self.inputs.add(node.attr)
            elif isinstance(node.ctx, ast.Store):
                self.outputs.add(node.attr)


def find_deps(node):
    """Find attributes of "self"."""
    fdv = _FindDepsVisitor()
    fdv.visit(node)
    return fdv



from shaderdef.material import create_stages, find_external_links
from shaderdef.stage import make_prefix


class ShaderDef(object):
    def __init__(self, material):
        self._material = material

        self._stages = []
        self._vert_shader = None
        self._frag_shader = None

    def _thread_deps(self):
        """Link inputs and outputs between stages."""
        iter1 = reversed(self._stages)
        iter2 = reversed(self._stages)
        next(iter2)
        for stage, prev_stage in zip(iter1, iter2):
            stage.input_prefix = make_prefix(prev_stage.name)
            prev_stage.provide_deps(stage)

    def translate(self):
        self._stages = list(create_stages(self._material))
        self._thread_deps()

        external_links = find_external_links(self._material)

        # TODO
        self._vert_shader = self._stages[0].to_glsl(external_links,
                                                    self._material.__class__)
        self._frag_shader = self._stages[1].to_glsl(external_links,
                                                    self._material.__class__)

    @property
    def vert_shader(self):
        if self._vert_shader is None:
            raise ValueError('material has not been translated yet')
        # TODO: declare inputs/outputs
        return self._vert_shader

    @property
    def frag_shader(self):
        if self._frag_shader is None:
            raise ValueError('material has not been translated yet')
        # TODO: declare inputs/outputs
        return self._frag_shader



from shaderdef.equality import EqualityMixin

class Attribute(EqualityMixin):
    def __init__(self, gtype):
        self.gtype = gtype

    def __repr__(self):
        return 'Attribute({})'.format(self.gtype)

    def glsl_decl(self, name):
        return 'in {} {};'.format(self.gtype.__name__, name)


class FragOutput(EqualityMixin):
    def __init__(self, gtype):
        self.gtype = gtype

    def __repr__(self):
        return 'FragOutput({})'.format(self.gtype)

    def glsl_decl(self, name):
        return 'out {} {};'.format(self.gtype.__name__, name)


class Uniform(EqualityMixin):
    def __init__(self, gtype):
        self.gtype = gtype

    def __repr__(self):
        return 'Uniform({})'.format(self.gtype)

    def glsl_decl(self, name):
        return 'uniform {} {};'.format(self.gtype.__name__, name)


class BuiltinType(object):
    pass


# pylint: disable=invalid-name
class mat2(BuiltinType):
    pass
class mat3(BuiltinType):
    pass
class mat4(BuiltinType):
    pass
class vec2(BuiltinType):
    pass
class vec3(BuiltinType):
    pass
class vec4(BuiltinType):
    pass
class void(BuiltinType):
    pass



from unittest import TestCase

from shaderdef.glsl_types import Uniform, vec3

class TestGlslTypes(TestCase):
    def test_uniform_decl(self):
        self.assertEqual(Uniform(vec3).glsl_decl('x'), 'uniform vec3 x;')
